@@ -12,6 +12,7 @@
 #include "key.h"
 #include "binding.h"
 #include "popup.h"
+#include "util.h"
 #include "xchainkeys.h"
 
 extern XChainKeys_t *xc;
@@ -21,8 +22,12 @@ Binding_t* binding_new() {
 
   self->key = NULL;
   self->action = XC_ACTION_ENTER;
-  self->argument = "";
 
+  self->name = calloc(128, sizeof(char));
+  strcpy(self->name, "default");
+
+  self->argument = calloc(4096, sizeof(char));
+  
   self->timeout = 3000;
   self->abort = XC_ABORT_AUTO;
 
@@ -40,6 +45,16 @@ void binding_set_action(Binding_t *self, char *str) {
       break;
     }
   }
+  if(strcmp(str, ":reload") == 0) {
+    self->action = XC_ACTION_LOAD;
+  }
+  if(strcmp(str, ":repeat") == 0) {
+    self->action = XC_ACTION_GROUP;
+    self->name = "default";
+    fprintf(stderr, 
+	    "%s: warning: ':repeat' is deprecated, using ':group \"default\"' instead.\n",
+	    PACKAGE_NAME);
+  }
 }
 
 void binding_parse_arguments(Binding_t *self) {
@@ -63,6 +78,7 @@ void binding_parse_arguments(Binding_t *self) {
     strncpy(argument, self->argument, strlen(self->argument));
 
     while(strlen(argument)) {
+
       if(strncmp(argument, "timeout=", 8) == 0) {
 
 	value = (char *) calloc(strlen(argument)+1, sizeof(char));
@@ -101,7 +117,6 @@ void binding_parse_arguments(Binding_t *self) {
 	free(value_ptr);
       }
 
-      argument+=strcspn(argument, ws);
       if(strcspn(argument, ws) == 0)
 	break;
     }
@@ -114,6 +129,42 @@ void binding_parse_arguments(Binding_t *self) {
   }
 }
 
+void binding_create_default_bindings(Binding_t *self) {
+
+  Binding_t *binding;
+  char *keyspec;
+  int i;
+
+  if(self->parent != NULL && self->action == XC_ACTION_ENTER) {
+
+    /* create default :escape binding unless present */
+    if (!binding_get_child_by_action(self, XC_ACTION_ESCAPE)) {
+      binding = binding_new();
+      binding->action = XC_ACTION_ESCAPE;
+
+      keyspec = key_to_str(self->key);
+      binding->key = key_new(keyspec);
+      free(keyspec);
+
+      binding_append_child(self, binding);
+    }
+    
+    /* create default :abort binding unless present */
+    if (!binding_get_child_by_action(self, XC_ACTION_ABORT)) {
+      binding = binding_new();
+      binding->action = XC_ACTION_ABORT;
+      binding->key = key_new("C-g");
+      binding_append_child(self, binding);
+    }
+  }
+  
+  /* recurse children */
+  for (i=0; i<self->num_children; i++) {
+    binding_create_default_bindings(self->children[i]);
+  }
+
+}
+
 void binding_append_child(Binding_t *self, Binding_t *child) {
   self->children[self->num_children] = child;
   child->parent = self;
@@ -160,12 +211,14 @@ void binding_activate(Binding_t *self) {
     binding_exec(self);
     break;
 
-  case XC_ACTION_REPEAT:
-    binding_repeat(self);
+  case XC_ACTION_GROUP:
+    binding_group(self);
     break;
 
-  case XC_ACTION_SEND:
-    binding_send(self);
+  case XC_ACTION_LOAD:
+    if(strlen(self->argument))
+      strcpy(xc->config, self->argument);
+    xc->reload = True;
     break;
   }
   free(path);
@@ -183,9 +236,9 @@ void binding_enter(Binding_t *self) {
   char *path = binding_to_path(self);
 
   /* prepare timeout and popup delay */
-  now = xc_get_msec();
+  now = get_msec();
   delay = now + xc->delay;
-  timeout = xc_get_msec() + self->timeout;
+  timeout = get_msec() + self->timeout;
 
   /* prepare popup */
   snprintf(xc->popup->text, strlen(path)+1, "%s", path);
@@ -200,7 +253,7 @@ void binding_enter(Binding_t *self) {
   }
 
   while(!done) {    
-    now = xc_get_msec();
+    now = get_msec();
 
     /* show popup after delay */
     if(now >= delay && !xc->popup->mapped)
@@ -228,13 +281,13 @@ void binding_enter(Binding_t *self) {
 	keystr = XKeysymToString(XKeycodeToKeysym(xc->display, keycode, 0));
 	
 	/* check if this key is a modifier */
-	if (xc_keycode_to_modmask(xc, keycode) != 0) {
+	if (keycode_to_modifier(xc->xmodmap, keycode) != 0) {
 	  continue;
 	}      
 	else {	
 	  /* non-modifier key hit... */
 	  key = key_new(keystr);
-	  key->modifiers = xc_get_modifiers(xc);
+	  key->modifiers = get_modifiers(xc->display);
 	  
 	  /* check if this key is bound in this keymap */
 	  if( (binding = binding_get_child_by_key(self, key)) != NULL) {
@@ -294,32 +347,13 @@ void binding_escape(Binding_t *self) {
   XUngrabKeyboard(xc->display, CurrentTime);
   XGetInputFocus(xc->display, &window, &focus_ret);
 
-  xc_send_key(xc, self->parent->key, window);
+  send_key(xc->display, self->parent->key, window);
 
   XGrabKeyboard(xc->display, DefaultRootWindow(xc->display),
 		True, GrabModeAsync, GrabModeAsync, CurrentTime);
 }
 
-void binding_send(Binding_t *self) {
-
-  Key_t *key;
-  Window root = DefaultRootWindow(xc->display);
-
-  key = key_new(self->argument);
-
-  if (key != NULL) {
-
-    XUngrabKeyboard(xc->display, CurrentTime);
-
-    xc_send_key(xc, key, root);
-
-    XGrabKeyboard(xc->display, DefaultRootWindow(xc->display),
-		  True, GrabModeAsync, GrabModeAsync, CurrentTime);
-    free(key);
-  }
-}
-
-void binding_repeat(Binding_t *self) {
+void binding_group(Binding_t *self) {
   Key_t *key;
   Binding_t *binding;
   XEvent event;
@@ -342,26 +376,27 @@ void binding_repeat(Binding_t *self) {
       keystr = XKeysymToString(XKeycodeToKeysym(xc->display, keycode, 0));
       
       /* check if this key is a modifier */
-      if (xc_keycode_to_modmask(xc, keycode) != 0) {
+      if (keycode_to_modifier(xc->xmodmap, keycode) != 0) {
 	break;
       }      
       else {	
 	/* non-modifier key hit... */
 	key = key_new(keystr);
-	key->modifiers = xc_get_modifiers(xc);
+	key->modifiers = get_modifiers(xc->display);
 	
-	/* :exec all :repeat actions in the parent binding 
-	 * abort on any non-repeating key
+	/* :exec any :group actions of the same name in the parent
+	 * binding, and abort on any non-repeating key
 	 */
 	for(i=0; i<self->parent->num_children; i++) {
 	  binding = self->parent->children[i];
+	  
+	  if ( binding->action == XC_ACTION_GROUP &&
+	       strcmp(binding->name, self->name) == 0 &&
+	       key_equals(binding->key, key) ) {
 
-	  if ( binding->action == XC_ACTION_REPEAT) { 
-	    if ( key_equals(binding->key, key) ) {
-	      binding_exec(binding);
-	      abort = False;
-	      break;
-	    }
+	    binding_exec(binding);
+	    abort = False;
+	    break;
 	  }
 	  abort = True;
 	}
@@ -439,8 +474,14 @@ void binding_list(Binding_t *self) {
 
   if(depth > 0) {
     keyspec = key_to_str(self->key);
-    printf("%s %s %s\n", 
-	    keyspec, xc->action_names[self->action], self->argument);
+    
+    if(strcmp(self->name, "default") == 0)
+      printf("%s %s %s\n", 
+	     keyspec, xc->action_names[self->action], self->argument);
+    else 
+      printf("%s %s \"%s\" %s\n", 
+	     keyspec, xc->action_names[self->action], self->name, self->argument);
+
     free(keyspec);
   }
 
@@ -448,3 +489,18 @@ void binding_list(Binding_t *self) {
     binding_list(self->children[i]);
   }
 }
+
+void binding_free(Binding_t *self) {
+  int i;
+
+  for( i=0; i<self->num_children; i++ ) {
+    binding_free(self->children[i]);
+    self->children[i] = NULL;
+  }
+  self->num_children = 0;
+  self->parent = NULL;
+  free(self->name);
+  free(self->argument);
+  free(self->key);
+  free(self);
+}
