@@ -18,6 +18,6 @@ void popup_update(Popup_t *self);
 void popup_show(Popup_t *self);
 void popup_hide(Popup_t *self);
 void popup_set_timeout(Popup_t *self, unsigned int ms);
-void popup_destroy(Popup_t *self);
+void popup_free(Popup_t *self);
 
 #endif
