@@ -30,87 +30,125 @@
 #include <string.h>
 #include <errno.h>
 #include <getopt.h>
-#include <sys/time.h>
 #include <X11/Xlib.h>
 
 #include "key.h"
 #include "binding.h"
 #include "popup.h"
+#include "util.h"
 #include "xchainkeys.h"
 
 XChainKeys_t *xc;
 
-XChainKeys_t* xc_new(Display *display) {
+XChainKeys_t* xc_new() {
 
-  XChainKeys_t *self;
-  unsigned int i, num, caps, scroll;
+  XChainKeys_t *self = (XChainKeys_t *) calloc(1, sizeof(XChainKeys_t)); 
 
-  self = (XChainKeys_t *) calloc(1, sizeof(XChainKeys_t)); 
+  if(NULL == (self->display=XOpenDisplay(NULL))) {
+    fprintf(stderr, "%s: error: XOpenDisplay() failed for DISPLAY=%s.\n", 
+	    PACKAGE_NAME, getenv("DISPLAY"));
+    free(self);
+    exit(EXIT_FAILURE);
+  }
 
-  self->display = display;
   self->debug = False;
   self->timeout = 3000;
   self->delay = 1000;
   self->reentry = NULL;
+  self->reload = False;
+
   self->xmodmap = XGetModifierMapping(self->display);
-  
+  xc_init_modmask(self);
+
   self->action_names[0] = ":none";
   self->action_names[1] = ":enter";
   self->action_names[2] = ":escape";
   self->action_names[3] = ":abort";
   self->action_names[4] = ":exec";
-  self->action_names[5] = ":repeat";
-  self->action_names[6] = ":send";
+  self->action_names[5] = ":group";
+  self->action_names[6] = ":load";
   self->num_actions = 7;
 
   self->root = binding_new();
   self->root->action = XC_ACTION_NONE;
   
-  /* prepare modmask array, containing all possible combinations of
-   * num, caps and scroll lock */
-
-  i = 0;
-  num = 
-    xc_keycode_to_modmask(self, XKeysymToKeycode(self->display, XStringToKeysym("Num_Lock")));
-  caps = 
-    xc_keycode_to_modmask(self, XKeysymToKeycode(self->display, XStringToKeysym("Caps_Lock")));
-  scroll = 
-    xc_keycode_to_modmask(self, XKeysymToKeycode(self->display, XStringToKeysym("Scoll_Lock")));
-
-  self->modmask[i++] = 0;
-  self->modmask[i++] = num;
-  self->modmask[i++] = caps;
-  self->modmask[i++] = scroll;
-  self->modmask[i++] = num | caps;
-  self->modmask[i++] = num | scroll;
-  self->modmask[i++] = caps | scroll;
-  self->modmask[i++] = num | caps | scroll;
-
   xc_find_config(self);
 
   return self;
 }
 
-void xc_version(XChainKeys_t *self) {
-  printf("%s %s Copyright (C) 2010 Henning Bekel <%s>\n",
-	 PACKAGE_NAME, PACKAGE_VERSION, PACKAGE_BUGREPORT);
+void xc_init_modmask(XChainKeys_t *self) {
+  unsigned int num, caps, scroll;
+
+  /* self->modmask contains all possible combinations of num, caps and
+   * scroll lock */
+
+  num = 
+    keycode_to_modifier(self->xmodmap, 
+			XKeysymToKeycode(self->display, XStringToKeysym("Num_Lock")));
+  caps = 
+    keycode_to_modifier(self->xmodmap, 
+			XKeysymToKeycode(self->display, XStringToKeysym("Caps_Lock")));
+  scroll = 
+    keycode_to_modifier(self->xmodmap, 
+			XKeysymToKeycode(self->display, XStringToKeysym("Scoll_Lock")));
+
+  self->modmask[0] = 0;
+  self->modmask[1] = num;
+  self->modmask[2] = caps;
+  self->modmask[3] = scroll;
+  self->modmask[4] = num | caps;
+  self->modmask[5] = num | scroll;
+  self->modmask[6] = caps | scroll;
+  self->modmask[7] = num | caps | scroll;
 }
 
-void xc_usage(XChainKeys_t *self) {
-  printf("Usage: %s [options]\n\n", PACKAGE_NAME);
-  printf("  -d, --debug   : Enable debug messages\n");
-  printf("  -h, --help    : Print this help text\n");
-  printf("  -v, --version : Print version information\n");
-  printf("  -k, --keys    : Show valid keyspecs\n");
-  printf("  -f, --file    : alternative config file\n");
-  printf("\n");
+void xc_show_keys(XChainKeys_t *self) {
+
+  XEvent event;
+  Key_t *key;
+  KeyCode keycode;
+  char *keystr;
+  char *keyspec;
+
+  version();
+  printf("\nPress a key combination to show the corresponding keyspec.\n");
+  printf("Press Control-c to quit.\n\n");
+  
+  XGrabKeyboard(self->display, DefaultRootWindow(self->display),
+		True, GrabModeAsync, GrabModeAsync, CurrentTime);
+
+  while(True) {
+    XNextEvent(self->display, &event);
+    switch(event.type) {
+    case KeyPress:
+      
+      keycode = ((XKeyPressedEvent*)&event)->keycode;
+      keystr = XKeysymToString(XKeycodeToKeysym(self->display, keycode, 0));
+      
+      if (keycode_to_modifier(xc->xmodmap, keycode) != 0) 
+	continue;
+      
+      if ( strcmp(keystr, "c") == 0 && get_modifiers(self->display) == ControlMask ) {
+	XUngrabKeyboard(self->display, CurrentTime);
+	return;
+      }
+      
+      key = key_new(keystr);      
+      key->modifiers = get_modifiers(self->display);
+      
+      keyspec = key_to_str(key);
+      printf("%s\n", keyspec);
+      free(keyspec);
+      free(key);
+    }
+  }
 }
 
 void xc_find_config(XChainKeys_t *self) {
   
   self->config = (char *) calloc(4096, sizeof(char));
 
-  /* determine config file self->config */
   strcpy(self->config, getenv("HOME"));
   
   if(getenv("XDG_CONFIG_HOME") != NULL ) {
@@ -127,14 +165,15 @@ void xc_parse_config(XChainKeys_t *self) {
   FILE *f;
   char *buffer = (char *) calloc(4096, sizeof(char));
   char *argument= (char *) calloc(4096, sizeof(char));
-  char *line, *token, *expect;
+  char *line, *token, *expect, *path;
   const char *ws = " \t"; 
   int linenum = 0;
-  int len, pos, i;
+  int len, pos;
 
   Key_t *key;
   Binding_t *binding;
   Binding_t *parent;
+  Binding_t *existing;
 
   int feedback = True;
 
@@ -208,7 +247,8 @@ void xc_parse_config(XChainKeys_t *self) {
       line += 4;
       line += strspn(line, ws);
       line[strcspn(line, ws)] = '\0';
-      font = strdup(line);
+      strncpy(font, line, strlen(line));
+      font[strlen(line)] = '\0';
       continue;
     }
 
@@ -216,7 +256,8 @@ void xc_parse_config(XChainKeys_t *self) {
       line += 10;
       line += strspn(line, ws);
       line[strcspn(line, ws)] = '\0';
-      fg = strdup(line);
+      strncpy(fg, line, strlen(line));
+      fg[strlen(line)] = '\0';
       continue;
     }
 
@@ -224,7 +265,8 @@ void xc_parse_config(XChainKeys_t *self) {
       line += 10;
       line += strspn(line, ws);
       line[strcspn(line, ws)] = '\0';
-      bg = strdup(line);
+      strncpy(bg, line, strlen(line));
+      bg[strlen(line)] = '\0';
       continue;
     }
 
@@ -254,10 +296,24 @@ void xc_parse_config(XChainKeys_t *self) {
 	  /* if this key is already bound in the parent binding,
 	   * then make that binding the parent binding for the next key
 	   */
-	  if(binding_get_child_by_key(parent, key) != NULL) {
-	    parent = binding_get_child_by_key(parent, key);
-	    goto next_token;
-	  }  
+	  existing = binding_get_child_by_key(parent, key);
+
+	  if( existing != NULL) {
+	    if(line[0] == ':' || line[0] == '\0') {
+	      path = binding_to_path(existing);
+	      fprintf(stderr, "%s: warning: line %d: %s: already bound, skipping...\n",
+		      PACKAGE_NAME, linenum, path);
+	      free(path);
+	      free(key);
+	      free(token);
+	      goto next_line;
+	    }
+	    else {
+	      parent = binding_get_child_by_key(parent, key);
+	      free(key);
+	      goto next_token;
+	    }
+	  }
 
 	  /* create a binding for this key */
 	  binding = binding_new();
@@ -271,19 +327,34 @@ void xc_parse_config(XChainKeys_t *self) {
 	}
 	else {
 	  fprintf(stderr,
-		  "%s: warning: Invalid keyspec: '%s'. Use %s -k to show valid keyspecs.\n",
-		  PACKAGE_NAME, token, PACKAGE_NAME);
-	  fprintf(stderr, "-> Skipping line %d: '%s'\n", linenum, buffer);
+		  "%s: warning: line %d: %s: invalid keyspec, skipping...\n",
+		  PACKAGE_NAME, linenum, token);
+	  free(token);
 	  goto next_line;
 	}
       }
       
       if (strcmp(expect, "action") == 0) {
 	binding_set_action(binding, token);
-	expect = "argument";
+	expect = "name";
 	goto next_token;
       }
 
+      if (strcmp(expect, "name") == 0) {
+	if( token[0] == '"' && token[strlen(token)-1] == '"' ) {
+	  token += 1; 
+	  token[strlen(token)-1] = '\0';
+
+	  strcpy(binding->name, token);
+	  token -= 1;
+	  expect = "argument";
+	  goto next_token;
+	}
+	else {
+	  expect = "argument";
+	}
+      }
+
       if (strcmp(expect, "argument") == 0) {
 	if(strlen(argument))
 	  strncat(argument, " ", 1);
@@ -300,7 +371,7 @@ void xc_parse_config(XChainKeys_t *self) {
     if (binding != NULL) {
       /* append the argument to the current binding */
       if (strlen(argument)) {
-	binding->argument = strdup(argument);
+	strcpy(binding->argument, argument);
       }
     }
   next_line:
@@ -308,31 +379,8 @@ void xc_parse_config(XChainKeys_t *self) {
   }
   fclose(f);
 
-  /* parse binding arguments */
-  binding_parse_arguments(xc->root);
-
-  /* add defaults */
-  for (i=0; i<self->root->num_children; i++) {
-    parent = self->root->children[i];
-    if (parent->action == XC_ACTION_ENTER) {
-
-      /* create default :escape binding unless present */
-      if (!binding_get_child_by_action(parent, XC_ACTION_ESCAPE)) {
-	binding = binding_new();
-	binding->action = XC_ACTION_ESCAPE;
-	binding->key = parent->key;
-	binding_append_child(parent, binding);
-      }
-
-      /* create default :abort binding unless present */
-      if (!binding_get_child_by_action(parent, XC_ACTION_ABORT)) {
-	binding = binding_new();
-	binding->action = XC_ACTION_ABORT;
-	binding->key = key_new("C-g");
-	binding_append_child(parent, binding);
-      }
-    }
-  }
+  binding_parse_arguments(self->root);
+  binding_create_default_bindings(self->root);
 
   /* initialize popup window */
   self->popup = popup_new(self->display, font, fg, bg);
@@ -358,12 +406,19 @@ void xc_parse_config(XChainKeys_t *self) {
   }
   free(buffer);
   free(argument);
-
   free(font);
   free(fg);
   free(bg);
 }
 
+void xc_grab_prefix_keys(XChainKeys_t *self) {
+  /* grab top level keys individually */
+  int i;
+  for (i=0; i<self->root->num_children; i++) {
+    key_grab(self->root->children[i]->key);    
+  }
+}
+
 void xc_mainloop(XChainKeys_t *self) {
   Binding_t *binding;
   Binding_t *reentry;
@@ -372,11 +427,7 @@ void xc_mainloop(XChainKeys_t *self) {
   long now;
   int i;
 
-  /* grab top level keys individually */
-  for (i=0; i<self->root->num_children; i++) {
-    binding = self->root->children[i];
-    key_grab(binding->key);    
-  }
+  xc_grab_prefix_keys(self);
 
   while(True) {
 
@@ -384,7 +435,7 @@ void xc_mainloop(XChainKeys_t *self) {
       if(xc->popup->timeout == 0)
 	break;
       
-      now = xc_get_msec();
+      now = get_msec();
       
       if(now >= xc->popup->timeout) {
 	popup_hide(xc->popup);
@@ -401,7 +452,7 @@ void xc_mainloop(XChainKeys_t *self) {
 	binding = self->root->children[i];
 	
 	if (binding->key->keycode == keycode) {
-	  if(binding->key->modifiers == xc_get_modifiers(xc)) {
+	  if(binding->key->modifiers == get_modifiers(xc->display)) {
 
 	    popup_hide(xc->popup);
 	    xc->popup->timeout = 0;
@@ -419,131 +470,40 @@ void xc_mainloop(XChainKeys_t *self) {
       binding_activate(reentry);
       goto reentry;
     }
-  }
-}
 
-void xc_show_keys(XChainKeys_t *self) {
-
-  XEvent event;
-  Key_t *key;
-  KeyCode keycode;
-  char *keystr;
-  char *keyspec;
-
-  xc_version(xc);
-  printf("\nPress a key combination to show the corresponding keyspec.\n");
-  printf("Press Control-c to quit.\n");
-  
-  XGrabKeyboard(self->display, DefaultRootWindow(self->display),
-		True, GrabModeAsync, GrabModeAsync, CurrentTime);
-
-  while(True) {
-    XNextEvent(self->display, &event);
-    switch(event.type) {
-    case KeyPress:
-      
-      keycode = ((XKeyPressedEvent*)&event)->keycode;
-      keystr = XKeysymToString(XKeycodeToKeysym(self->display, keycode, 0));
-      
-      if (xc_keycode_to_modmask(xc, keycode) != 0) 
-	continue;
-      
-      if ( strcmp(keystr, "c") == 0 && xc_get_modifiers(self) == ControlMask ) {
-	XUngrabKeyboard(self->display, CurrentTime);
-	return;
-      }
-      
-      key = key_new(keystr);      
-      key->modifiers = xc_get_modifiers(self);
-      
-      keyspec = key_to_str(key);
-      printf("%s\n", keyspec);
-      free(keyspec);
-      free(key);
+    if(xc->reload) {
+      xc_reload(self);
     }
   }
 }
 
-void xc_send_key(XChainKeys_t *self, Key_t *key, Window window) {
+void xc_reset(XChainKeys_t *self) {
+  int i;
+  Binding_t *prefix;
 
-  XKeyEvent e;
-  char *keyspec;
+  for(i=0; i<self->root->num_children; i++) {
+    prefix = self->root->children[i];
 
-  if (self->debug) {
-    keyspec = key_to_str(key);
-      printf("Sending synthetic KeyPressEvent (%s) to window id %d\n", 
-	      keyspec, (int)window);
-    free(keyspec);
+    key_ungrab(prefix->key);
+    binding_free(prefix);
+    self->root->children[i] = NULL;
   }
+  self->root->num_children = 0;
 
-  e.display = self->display;
-  e.window = window;
-  e.subwindow = None;
-  e.time = CurrentTime;
-  e.same_screen = True;	   
-  e.keycode = key->keycode;
-  e.state = key->modifiers;
-  e.type = KeyPress;
-  e.x = e.y = e.x_root = e.y_root = 1;
-  
-  XSendEvent(self->display, e.window, True, KeyPressMask, (XEvent *)&e);
-  XSync(self->display, False);
+  self->reentry = NULL;
+  self->reload = False;
+
+  popup_free(self->popup);
+  self->popup = NULL;
 }
 
-int xc_keycode_to_modmask(XChainKeys_t *self, KeyCode keycode) {
-
-  int i = 0;
-  int j = 0;
-  int max = self->xmodmap->max_keypermod;
-
-  for (i = 0; i < 8; i++) {
-    for (j = 0; j < max && self->xmodmap->modifiermap[(i * max) + j]; j++) {
-      if (keycode == self->xmodmap->modifiermap[(i * max) + j]) {
-        switch (i) {
-          case ShiftMapIndex: return ShiftMask; break;
-          case LockMapIndex: return LockMask; break;
-          case ControlMapIndex: return ControlMask; break;
-          case Mod1MapIndex: return Mod1Mask; break;
-          case Mod2MapIndex: return Mod2Mask; break;
-          case Mod3MapIndex: return Mod3Mask; break;
-          case Mod4MapIndex: return Mod4Mask; break;
-          case Mod5MapIndex: return Mod5Mask; break;
-        }
-      } 
-    } 
-  } 
-  return 0;
+void xc_reload(XChainKeys_t *self) {  
+  xc_reset(self);
+  xc_parse_config(self);
+  xc_grab_prefix_keys(self);
 }
 
-unsigned int xc_get_modifiers(XChainKeys_t *self) {
-
-  char keymap[32]; 
-  unsigned int keycode;
-  unsigned int modifiers = 0;
-
-  XQueryKeymap(self->display, keymap);
-
-  for (keycode = 0; keycode < 256; keycode++) {
-    if ((keymap[(keycode / 8)] & (1 << (keycode % 8))) \
-        && xc_keycode_to_modmask(self, keycode)) {
-
-      modifiers |= xc_keycode_to_modmask(self, keycode);
-    }
-  }
-  return modifiers;
-}
-
-long xc_get_msec() {
-  long msec;
-  struct timeval tv;
-  gettimeofday(&tv, NULL);
-  msec = (long)(tv.tv_sec*1000 + (tv.tv_usec/1000));
-  return msec;
-}
-
-int main(int argc, char **argv) {
-
-  Display *display;
+void xc_parse_options(XChainKeys_t *self, int argc, char **argv) {
 
   struct option options[] = {
     { "debug", no_argument, NULL, 'd' },
@@ -555,16 +515,6 @@ int main(int argc, char **argv) {
   };
   int option, option_index;
 
-  /* try to open the x display */
-  if (NULL==(display=XOpenDisplay(NULL))) {
-    perror(argv[0]);
-    exit(EXIT_FAILURE);
-  }
-
-  /* create XChainkeys instance */
-  xc = xc_new(display);
- 
-  /* parse command line arguments */
   while (1) {
 
     option = getopt_long(argc, argv, "dhvkf:", options, &option_index);
@@ -572,26 +522,27 @@ int main(int argc, char **argv) {
     switch (option) {
 
     case 'f':
-      strncpy(xc->config, optarg, strlen(optarg));
-      xc->config[strlen(optarg)] = '\0';
+      strncpy(self->config, optarg, strlen(optarg));
+      self->config[strlen(optarg)] = '\0';
       break;      
 
     case 'k':
-      xc_show_keys(xc);
+      xc_show_keys(self);
       exit(EXIT_SUCCESS);
 
     case 'd':
-      xc->debug = True;
-      xc_version(xc);
+      self->debug = True;
+      version();
+      printf("\n");
       break;
       
     case 'h':
-      xc_usage(xc);
+      usage();
       exit(EXIT_SUCCESS);
       break;
       
     case 'v':
-      xc_version(xc);
+      version();
       exit(EXIT_SUCCESS);
       break;      
 
@@ -599,20 +550,22 @@ int main(int argc, char **argv) {
       break;
 
     default:
-      xc_usage(xc);
+      usage();
       exit(EXIT_FAILURE);
       break;
     }
     if(option == -1)
       break;
-  }
+  }  
+}
 
-  /* parse config file */
+int main(int argc, char **argv) {
+
+  xc = xc_new();
+
+  xc_parse_options(xc, argc, argv);
   xc_parse_config(xc);
-
-  /* enter mainloop */
   xc_mainloop(xc);
 
   exit(EXIT_SUCCESS);
 }
-
