@@ -11,29 +11,30 @@ typedef struct XChainKeys {
   unsigned int timeout;
   unsigned int delay;
   char *config;
+  int reload;
   struct Popup *popup;
   struct Binding *root;
   struct Binding *reentry;
 } XChainKeys_t;
 
-XChainKeys_t* xc_new(Display *display);
-void xc_version(XChainKeys_t *self);
-void xc_usage(XChainKeys_t *self);
+XChainKeys_t* xc_new(void);
+void xc_parse_options(XChainKeys_t *self, int argc, char **argv);
+void xc_init_modmask(XChainKeys_t *self);
+void xc_show_keys(XChainKeys_t *self);
 void xc_find_config(XChainKeys_t *self);
 void xc_parse_config(XChainKeys_t *self);
+void xc_grab_prefix_keys(XChainKeys_t *self);
 void xc_mainloop(XChainKeys_t *self);
-void xc_send_key(XChainKeys_t *self, Key_t *key, Window window);
-int xc_keycode_to_modmask(XChainKeys_t *self, KeyCode keycode);
-unsigned int xc_get_modifiers(XChainKeys_t *self);
-long xc_get_msec();
+void xc_reload(XChainKeys_t *self);
+void xc_reset(XChainKeys_t *self);
 
 #define XC_ACTION_NONE   0
 #define XC_ACTION_ENTER  1
 #define XC_ACTION_ESCAPE 2
 #define XC_ACTION_ABORT  3
 #define XC_ACTION_EXEC   4
-#define XC_ACTION_REPEAT 5
-#define XC_ACTION_SEND   6
+#define XC_ACTION_GROUP  5
+#define XC_ACTION_LOAD   6
 
 #define XC_ABORT_AUTO 1
 #define XC_ABORT_MANUAL 0
